# -*- coding: utf-8 -*-
from __future__ import division

from numpy import sum, sqrt, array, zeros, exp, log10, load, pi


elementaryCharge = 1.602176634e-19 # in Coulomb.
eps0 = 8.85418782e-12 # Permittivity of free space, in units of m^{-3} kg^{-1} s^2 C^2

def logLambda(ne,Te):
    """Calculates the electron-electron logarithm, using the same calculation as in the SFINCS code [https://github.com/landreman/sfincs] for the sake of comparison.
 Source might be: reference 6 in Braginskii 1965, Spitzer 'Physics of fully ionized gases' ??

    Inputs
     ne: electron density, in m^(-3)
     Te: electron temperature, in J
    Note
     The inputs are converted to units of 10^20 m^{-3} and keV for the calculation

    Output
     The Coulomb logarithm.
    """
    # Unit conversion
    ne=ne/1e20
    Te=Te/1.60217662e-16
    return 25.3 - 1.15*log10(ne*(1e14)) + (2.30e+0)*log10(Te*1000)

def FSA(X,B2):
    """Calculates the flux-surface average of X. 
    *Assumes Boozer coordinates.*
     
    Inputs
     X: 2D numpy array of a quantity which varies on the flux surface, sampled on an equidistant grid in Boozer-coordinates θ, ζ.
     B2: 2D numpy array of the magnetic field strength squared (B^2) on the flux-surface, sampled on the same grid as X.

    *NOTE*
     This is the only place where Boozer coordinates are assumed in the code. Substituting this function thus enables the usage of more general coordinates.
    """
    return sum(X/B2,axis=(-1,-2))/sum(1/B2)

def classical_transport(Zs, ms, Ns, Ts, dN_dpsis, dT_dpsis, logLambda, nablaPsi2, B2, tildePhi):
    """Calculates the classical transport given masses, charge and profiles of all the species, and a geometric factor |\nabla \psi|^2. 

    Inputs
     Zs : array of charge numbers for the different species. See note on '1D inputs'.
     ms: array of masses (in kilogram) for the different species. See note on '1D inputs'.
     Ns: array of pseudo-densities (in meter^{-3}) for the different species. See note on '1D inputs'.
     Ts: array of temperatures (in Joule) for the different species. See note on '1D inputs'.
     dN_dpsis: dN/dψ (in meter^{-3}/[ψ]), radial derivative of pseudo-densities for the different species. See notes on radial coordinates below. Also see note on '1D inputs'.
     dT_dpsis: dT/dψ (in Joule^{-3}/[ψ]). radial derivative of temperature for the different species. See notes on radial coordinates below. Also see note on '1D inputs'.
     logLambda: The Coulomb logarithm.
     nablaPsi2: |∇ψ|^2, see note on radial coordinates below.
     B2: the magnetic field strength squared (B^2) on the flux-surface, sampled on an equidistant 2D grid in Boozer coordinates theta and zeta. See note on '2D inputs'.
     tildePhi: the variation of the electrostatic potential on the flux-surface (in Volt). See note on '2D inputs'.

    Outputs
     classicalPF: the (normalized) radial classical particle flux, <Γ • ∇ψ>. 
     classicalHF: the (normalized) radial classical particle flux, <Q • ∇ψ>. 
     
    
    Radial coordinate:
     The gradients of density and temperature, and |∇ψ|^2, should use the same radial coordinate ψ. This ψ can be any flux-surface label. The calculated radial fluxes output by this function will be
     <Γ • ∇ψ>
     <Q • ∇ψ>
     where Γ and Q are the (vector) classical fluxes of particles and kinetic energy. Thus, the output fluxes (and even which units they are given in) are affected by the choice of radial coordinate in the inputs.

    1D inputs:
     These inputs are numpy arrays, with each entry corresponding to a species. The species can be given in any order, but the different arrays must use the same order, i.e. Zs[0], ms[0] should be the charge number and mass number of 'species 0', which can be any species (electron, hydrogen, an impurity, etc.).

    2D inputs:
     These are quantities that vary on the flux surface, sampled on an equidistant grid in Boozer-coordinates θ, ζ. If the grid has Nθ points in the θ coordinate and Nζ points in the ζ coordinate, the input should be a 2D numpy array with shape (Nθ, Nζ) or (Nζ, Nθ). The shape should be same for each 2D array.
    """
    
    # Get the number of species from the number of masses:
    Nspecies = len(ms)
    # Check that the lenght of each 1 dimensional array is the same:
    # (If not, an error will be raised and the calculation will stop.)
    assert(all([len(x) == Nspecies for x in [Zs,Ns,Ts,dN_dpsis,dT_dpsis]]))

    # For each species, calculate the density variation on the flux-surface
    # from the pseudo-density
    ns = zeros((Nspecies,) + tildePhi.shape)
    for i in range(Nspecies):
        ns[i] = Ns[i] * exp(-Zs[i]*elementaryCharge*tildePhi/Ts[i])
            

    # classical particle flux:
    classicalPF = zeros(Nspecies)
    # classical kinetic energy (heat) flux:
    classicalHF = zeros(Nspecies)
    
    for a in range(0,Nspecies):
        for b in range(0,Nspecies):

            # Ratio of squared thermal speed of species b and a:
            xab2 = ms[a]*Ts[b]/(ms[b]*Ts[a])
            # Braginskii matrix elements:
            Mab00 = - (1+ms[a]/ms[b]) * (1 + xab2)
            Mab01 = - 1.5 * (1+ms[a]/ms[b])
            Mab11 = -(13. + 16. * xab2 + 30. * xab2**2)/4.
            Nab11 = 27. * ms[a]/(4.*ms[b])
            Mab00 = Mab00/((1+xab2)**(2.5))
            Mab01 = Mab01/((1+xab2)**(2.5))
            Mab11 = Mab11/((1+xab2)**(2.5))
            Nab11 = Nab11/((1+xab2)**(2.5))

            # Geometric prefactors:
            G1ab =  FSA(nablaPsi2 * ns[a] * ns[b]/B2,B2)
            G2ab =  FSA(nablaPsi2 * ns[a] * ns[b] *elementaryCharge*tildePhi/B2,B2)

            # Classical particle flux:
            classicalPF[a] = classicalPF[a] \
                             + Zs[b]**2  *( \
                                            G1ab * Mab00 * (Ts[a] * dN_dpsis[a]/(Ns[a] * Zs[a]) - Ts[b] * dN_dpsis[b]/(Ns[b] * Zs[b])) \
                                            + G2ab * Mab00 * ( dT_dpsis[a]/Ts[a] -  dT_dpsis[b]/Ts[b]) \
                                            + G1ab * ((Mab00 - Mab01) *  dT_dpsis[a]/Zs[a] - (Mab00 - xab2*Mab01) *  dT_dpsis[b]/Zs[b]) \
                                        )

            # Classical heat flux:
            # This only gives the conductive heat flux, typically denoted with small q,
            # the convective part is added below to make it the total heat flux, big Q.
            classicalHF[a] = classicalHF[a] \
                             + Zs[b]**2 * ( \
                                                      G1ab * Mab01 * (Ts[a] * dN_dpsis[a]/(Ns[a] * Zs[a]) - Ts[b] * dN_dpsis[b]/(Ns[b] * Zs[b])) \
                                                      + G2ab * Mab01 * ( dT_dpsis[a]/Ts[a] -  dT_dpsis[b]/Ts[b]) \
                                                      + G1ab * ((Mab01 - Mab11) *  dT_dpsis[a]/Zs[a] - (Mab01 + Nab11) *  dT_dpsis[b]/Zs[b]) \
                                                  )

        # Multiply with prefactors to get the fluxes in SI units
        classicalPF[a] = Zs[a] * sqrt(2) * elementaryCharge**2 * logLambda * sqrt(ms[a]) * classicalPF[a]/(12*pi**(3/2) * eps0**2 *Ts[a]**1.5)
        classicalHF[a] = -Zs[a] * sqrt(2) * elementaryCharge**2 * logLambda * sqrt(ms[a]) * classicalHF[a]/(12*pi**(3/2) * eps0**2 *sqrt(Ts[a]))
        # Add the convective heat flux to the conductive heat flux for the total heat flux
        classicalHF[a] = classicalHF[a] + 2.5 * Ts[a] * classicalPF[a]
    
    return classicalPF,classicalHF

if __name__ == "__main__":
    # B, tildePhi and |∇ψ|^2 data from 1 field period in a Wendelstein 7-X high-mirror case for rN=0.7865. The shape is (Nθ, Nζ), where Nθ, Nζ is the number of grid-points in the θ,ζ Boozer angle.
    B2 = load("B2.npy")
    # |∇ψ|^2 in units of (Tesla meter)^2, ψ is the toroidal flux in units of Tm^2
    nablaPsi2 = load("nablaPsi2.npy")
    # electrostatic potential variation in Volt
    tildePhi = load("tildePhi.npy")
    # To test calculation without tildePhi, just set it to a zero array:
    #tildePhi = array([0])

    # Gradients, mass, etc., for each species in the Wendelstein 7-X high-mirror case for rN=0.7865
    # Everything in SI units, ψ in Tm^2
    # Species:  electron, hydrogen, neon 10+
    dN_dpsis = array([-3.2513, -2.89005, -0.0361256]) * 1e20 # pseudo-density radial gradient, in 1/(Tm^5)
    Ns = array([0.652396013607, 0.579907567651, 0.00724884459563]) * 1e20 # pseudo-densities, in 1/m^3
    dT_dpsis = array([-7.27036, -8.57164, -8.57164]) * 1.60217662e-16 # temperature radial gradient, in J/(Tm^2)
    Ts = array([1.85395666892, 1.87954337045, 1.87954337045]) * 1.60217662e-16 #temperature, in J
    Zs = array([-1, 1, 10]) # charge number, dimensionless
    ms = array([0.000545509, 1.0, 20.18]) * 1.6726219e-27 # masses, in kg

    # Coulomb logarithm:
    lnLambda = logLambda(Ns[0],Ts[0]) # 16.92994103938514

    # Classical fluxes in SI units
    # classicalPF : Particle flux, in Tesla/(second meter)
    # classicalHF : Heat flux, in Joule Tesla/(second meter)
    classicalPF,classicalHF = classical_transport(Zs, ms, Ns, Ts, dN_dpsis, dT_dpsis, lnLambda, nablaPsi2, B2, tildePhi)

    print("Particle flux:")
    print(classicalPF)
    # With tildePhi ≠ 0:
    # [ 1.13350717e+17  7.03464910e+17 -5.90114193e+16]
    # With tildePhi = 0:
    # [ 1.13350731e+17  7.03465276e+17 -5.90114545e+16]
    
    print("Heat flux:")
    print(classicalHF)
    # # With tildePhi ≠ 0:
    # [  62.39844444 1238.59167865   10.91988073]
    # With tildePhi = 0:
    # [  62.39845205 1238.5921065    10.91990271
    
